/**
 * @file couple_config.c
 * @brief Structs and interfaces to defined coupling configuration
 *
 * Stores reads in coupling configuration and sets up data structures
 * to do the coupling.
 *
 * @copyright Copyright  (C)  2021 Moritz Hanke <hanke@dkrz.de>
 *                                 Rene Redler <rene.redler@mpimet.mpg.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler <rene.redler@mpimet.mpg.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef COUPLE_CONFIG_H
#define COUPLE_CONFIG_H

#include "mtime_calendar.h"
#include "interp_method_avg.h"
#include "interp_method_conserv.h"
#include "interp_method_nnn.h"
#include "interp_method_spmap.h"

#define MAX_ROUTINE_NAME_LENGTH (256)
#define MAX_FILE_NAME_LENGTH (512)

enum yac_time_unit_type {
   C_MILLISECOND = 0,
   C_SECOND      = 1,
   C_MINUTE      = 2,
   C_HOUR        = 3,
   C_DAY         = 4,
   C_MONTH       = 5,
   C_YEAR        = 6,
   C_ISO_FORMAT  = 7,
};

enum yac_reduction_type {
  INSTANT      = 0,
  ACCUMULATE   = 1,
  TIME_AVERAGE = 2,
};

enum yac_interpolation_list {
   UNDEFINED             = 0,
   AVERAGE               = 1,
   N_NEAREST_NEIGHBOR    = 2,
   CONSERVATIVE          = 3,
   SOURCE_TO_TARGET_MAP  = 4,
   FIXED_VALUE           = 5,
   USER_FILE             = 6,
   CHECK                 = 7,
   BERNSTEIN_BEZIER      = 8,
   RADIAL_BASIS_FUNCTION = 9,
   CREEP                 = 10,
};

union yac_interp_stack_config_entry {
  // - every interpolation type needs to have the entry type as first
  //   entry in order to be able to compare them using memcmp
  // - no pointers are allowed here in order to be comparable
  struct {
    enum yac_interpolation_list type;
  } general;
  struct {
    enum yac_interpolation_list type;
    enum yac_interp_avg_weight_type reduction_type;
    int partial_coverage;
  } average;
  struct {
    enum yac_interpolation_list type;
    struct yac_nnn_config config;
  } n_nearest_neighbor,
    radial_basis_function;
  struct {
    enum yac_interpolation_list type;
    int order;
    int enforced_conserv;
    int partial_coverage;
    enum yac_interp_method_conserv_normalisation normalisation;
  } conservative;
  struct {
    enum yac_interpolation_list type;
    double spread_distance;
    enum yac_interp_spmap_weight_type weight_type;
  } spmap;
  struct {
    enum yac_interpolation_list type;
  } hcsbb;
  struct {
    enum yac_interpolation_list type;
    char filename[MAX_FILE_NAME_LENGTH];
    char src_grid_name[MAX_FILE_NAME_LENGTH];
    char tgt_grid_name[MAX_FILE_NAME_LENGTH];
    int io_proc_step;
  } user_file;
  struct {
    enum yac_interpolation_list type;
    double value;
  } fixed;
  struct {
    enum yac_interpolation_list type;
    char constructor_key[MAX_ROUTINE_NAME_LENGTH];
    char do_search_key[MAX_ROUTINE_NAME_LENGTH];
  } check;
  struct {
    enum yac_interpolation_list type;
    int creep_distance;
  } creep;
};

struct yac_interp_stack_config {
  union yac_interp_stack_config_entry * config;
  size_t size;
};

struct yac_couple_config_grid {
  char const * name;
  int id;
};

struct yac_couple_config_transient {
  char const * name;
  int id;
  size_t collection_size;
};

struct yac_couple_config_component_transient {
  size_t transient_idx;
  size_t grid_idx;
};

struct yac_couple_config_component {

  struct yac_couple_config_component_transient * component_transients;
  size_t num_component_transients;

  char const * name;
  int id;
};

struct yac_couple_config_transient_couple {

  struct {
    size_t component_idx;
    size_t component_transient_idx;

    char const * timestep;
    int lag;
  } source, target;

  int mapping_on_source;

  struct yac_interp_stack_config interp_stack;

  enum yac_reduction_type coupling_period_operation;
  char const * coupling_period;

  int enforce_write_weight_file;
  char const * weight_file_name;
};

struct yac_couple_config_couple {

  size_t component_indices[2];

  struct yac_couple_config_transient_couple * transient_couples;
  size_t num_transient_couples;
};

struct yac_couple_config {

  enum yac_time_unit_type timestep_unit;
  char const * start_datetime;
  char const * end_datetime;
  calendarType calendar;

  struct yac_couple_config_grid * grids;
  size_t num_grids;

  struct yac_couple_config_component * components;
  size_t num_components;

  struct yac_couple_config_transient * transients;
  size_t num_transients;

  struct yac_couple_config_couple * couples;
  size_t num_couples;

  int redirect_stdout;
};

void yac_couple_config_delete(struct yac_couple_config * couple_config);
size_t yac_couple_config_get_num_couples(
  struct yac_couple_config * couple_config);
size_t yac_couple_config_get_num_couple_transients(
  struct yac_couple_config * couple_config, size_t couple_idx);
void yac_couple_config_get_couple_component_ids(
  struct yac_couple_config * couple_config, size_t couple_idx,
  int couple_component_ids[2]);
int yac_couple_config_component_name_is_valid(
  struct yac_couple_config * couple_config, char const * component_name);
size_t yac_couple_config_get_num_components(
  struct yac_couple_config * couple_config);
int yac_couple_config_get_component_id(
  struct yac_couple_config * couple_config, size_t component_idx);
char const * yac_couple_config_get_component_name(
  struct yac_couple_config * couple_config, size_t component_idx);
char const * yac_couple_config_get_component_name_by_id(
  struct yac_couple_config * couple_config, int component_id);
int yac_couple_config_get_component_id_by_name(
  struct yac_couple_config * couple_config, char const * component_name);
struct yac_interp_stack_config yac_couple_config_get_interp_stack(
  struct yac_couple_config * couple_config,
  size_t couple_idx, size_t transient_couple_idx);
void yac_couple_config_get_transient_grid_names(
  struct yac_couple_config * couple_config,
  size_t couple_idx, size_t transient_couple_idx,
  char const ** src_grid_name, char const ** tgt_grid_name);
void yac_couple_config_get_transient_couple_component_ids(
  struct yac_couple_config * couple_config,
  size_t couple_idx, size_t transient_couple_idx,
  int * src_component_id, int * tgt_component_id);
size_t yac_couple_config_get_transient_collection_size(
  struct yac_couple_config * couple_config,
  size_t couple_idx, size_t transient_couple_idx);
char const * yac_couple_config_get_transient_name(
  struct yac_couple_config * couple_config,
  size_t couple_idx, size_t transient_couple_idx);
int yac_couple_config_mapping_on_source(
  struct yac_couple_config * couple_config,
  size_t couple_idx, size_t transient_couple_idx);
enum yac_time_unit_type yac_couple_config_get_timestep_unit(
  struct yac_couple_config * couple_config);
char const * yac_couple_config_get_coupling_period(
  struct yac_couple_config * couple_config,
  size_t couple_idx, size_t transient_couple_idx);
void yac_couple_config_get_source_timestep(
  struct yac_couple_config * couple_config,
  size_t couple_idx, size_t transient_couple_idx, char const ** timestep, int * lag);
void yac_couple_config_get_target_timestep(
  struct yac_couple_config * couple_config,
  size_t couple_idx, size_t transient_couple_idx, char const ** timestep, int * lag);
enum yac_reduction_type yac_couple_config_get_coupling_period_operation(
  struct yac_couple_config * couple_config,
  size_t couple_idx, size_t transient_couple_idx);
void yac_couple_config_set_datetime(
  struct yac_couple_config * couple_config,
  char const * start, char const * end);
char const * yac_couple_config_get_start_datetime(
  struct yac_couple_config * couple_config);
char const * yac_couple_config_get_end_datetime(
  struct yac_couple_config * couple_config);
int yac_couple_config_get_grid_id_by_name(
  struct yac_couple_config * couple_config, char const * grid_name);
int yac_couple_config_enforce_write_weight_file(
  struct yac_couple_config * couple_config,
  size_t couple_idx, size_t transient_couple_idx);
char const * yac_couple_config_get_weight_file_name(
  struct yac_couple_config * couple_config,
  size_t couple_idx, size_t transient_couple_idx);
int yac_couple_config_redirect_stdout(
  struct yac_couple_config * couple_config);
int yac_couple_config_contains_grid_name(
  struct yac_couple_config * couple_config, char const * grid_name);
int yac_couple_config_get_transient_id_by_name(
  struct yac_couple_config * couple_config, char const * transient_name);
int yac_couple_config_transient_config_is_valid(
  struct yac_couple_config * couple_config,
  int component_id, int grid_id, int transient_id);
calendarType yac_couple_config_get_calendar(
  struct yac_couple_config * couple_config);

int yac_interp_stack_config_entry_compare(void const * a, void const * b);
int yac_interp_stack_config_compare(void const * a, void const * b);
struct interp_method ** yac_interp_stack_config_generate(
  struct yac_interp_stack_config interp_stack);

#endif // COUPLE_CONFIG_H
